import unittest
import warnings

import pandas as pd

from lhcsmapi.reference.Reference import Reference


class TestReference(unittest.TestCase):

    def test_get_quench_heater_reference_discharges_RB(self):
        # arrange
        cell = "A8R1"
        circuit_type = 'RB'

        # act
        ref_discharge_df = Reference.read_quench_heater_reference_discharges(circuit_type)
        timestamp_act = ref_discharge_df[ref_discharge_df["Cell"] == cell]["Timestamp"].values[0]

        # assert
        timestamp_exp = 1417014253613000000
        self.assertEqual(timestamp_exp, timestamp_act)

    def test_get_quench_heater_reference_discharges_RQ(self):
        # arrange
        cell = "16L2"
        circuit_type = 'RQ'

        # act
        ref_discharge_df = Reference.read_quench_heater_reference_discharges(circuit_type)
        timestamp_act = ref_discharge_df[ref_discharge_df["Cell"] == cell]["Timestamp"].values[0]

        # assert
        timestamp_exp = 1417079205409000000
        self.assertEqual(timestamp_exp, timestamp_act)

    def test_get_quench_heater_reference_discharge_RQ(self):
        # arrange
        cell = "16L2"
        circuit_type = 'RQ'

        # act
        timestamp_act = Reference.get_quench_heater_reference_discharge(circuit_type, cell)

        # assert
        timestamp_exp = 1417079205409000000
        self.assertEqual(timestamp_exp, timestamp_act)

    def test_get_quench_heater_reference_discharge_error(self):
        # arrange
        cell = ""
        circuit_type = '600A'

        # act
        with self.assertRaises(ValueError) as context:
            Reference.get_quench_heater_reference_discharge(circuit_type, cell)

        # assert
        self.assertEqual('Circuit type 600A not supported!', str(context.exception))

    def test_read_energy_extraction_reference_features_rb(self):
        # arrange
        circuit_type = 'RB'

        # act
        features_df_act = Reference.read_energy_extraction_reference_features(circuit_type)

        # assert
        features_df_ref = pd.DataFrame({'min': {'R_odd': 0.0675, 'R_even': 0.0675, 't_delay_ee_odd': 0.05,
                                                't_delay_ee_even': 0.55, 'tau_u_dump_res_odd': 110.0,
                                                'tau_u_dump_res_even': 110.0},
                                        'max': {'R_odd': 0.0825, 'R_even': 0.0825, 't_delay_ee_odd': 0.15,
                                                't_delay_ee_even': 0.65, 'tau_u_dump_res_odd': 130.0,
                                                'tau_u_dump_res_even': 130.0}})
        pd.testing.assert_frame_equal(features_df_ref, features_df_act)

    def test_read_energy_extraction_reference_features_error(self):
        # arrange
        circuit_type = '600A'

        # act
        with self.assertRaises(KeyError) as context:
            Reference.read_energy_extraction_reference_features(circuit_type)

        # assert
        self.assertEqual('\'Circuit type 600A not supported!\'', str(context.exception))

    def test_read_power_converter_reference_features_rb(self):
        # arrange
        circuit_type = 'RB'

        # act
        features_df_act = Reference.read_power_converter_reference_features(circuit_type)

        # assert
        features_df_ref = pd.DataFrame({'min': {'tau_i_meas': 90}, 'max': {'tau_i_meas': 110}})
        pd.testing.assert_frame_equal(features_df_ref, features_df_act)

    def test_read_power_converter_reference_features_error(self):
        # arrange
        circuit_type = 'IT'

        # act
        with self.assertRaises(ValueError) as context:
            Reference.read_power_converter_reference_features(circuit_type)

        # assert
        self.assertEqual('Reference features for IT circuit type are not defined.', str(context.exception))

    def test_get_power_converter_reference_fpa_rb(self):
        # arrange
        circuit_type = 'RB'
        circuit_name = 'RB.A12'

        # act
        fgc_pm_act = Reference.get_power_converter_reference_fpa(circuit_type, circuit_name, col='fgcPm')

        # assert
        fgc_pm_ref = '2018-03-17 16:00:28.680'
        self.assertEqual(fgc_pm_ref, fgc_pm_act)

    def test_get_power_converter_reference_fpa_rq(self):
        # arrange
        circuit_type = 'RQ'
        circuit_name = 'RQD.A12'

        # act
        fgc_pm_act = Reference.get_power_converter_reference_fpa(circuit_type, circuit_name, col='fgcPm')

        # assert
        fgc_pm_ref = '2018-03-18 11:58:33.600'
        self.assertEqual(fgc_pm_ref, fgc_pm_act)

    def test_get_power_converter_reference_fpa_it_error(self):
        # arrange
        circuit_type = 'IT'
        circuit_name = 'RQX.L1'

        # act
        with self.assertRaises(ValueError) as context:
            Reference.get_power_converter_reference_fpa(circuit_type, circuit_name, col='fgcPm')

        # assert
        self.assertEqual('Circuit type IT not supported!', str(context.exception))

    def test_compare_features_to_reference_error(self):
        # arrange
        features_df = pd.DataFrame()
        circuit_type = 'RB'
        system = 'VF'

        # act
        with self.assertRaises(KeyError) as context:
            Reference.compare_features_to_reference(features_df, circuit_type, system)

        # assert
        self.assertEqual('\'System VF is not supported!\'', str(context.exception))

    def test_compare_difference_of_features_to_reference_error(self):
        # arrange
        features_df = pd.DataFrame()
        reference_df = pd.DataFrame()
        circuit_type = 'RB'
        system = 'EE'
        wildcard = ''

        # act
        with self.assertRaises(KeyError) as context:
            Reference.compare_difference_of_features_to_reference(features_df, reference_df, circuit_type,
                                                                  system, wildcard)

        # assert
        self.assertEqual('\'System EE is not supported!\'', str(context.exception))

    def test_get_mp3_period(self):
        # arrange
        date_time = '2018-01-01 00:00:00.000+01:00'

        # act
        period_act = Reference.get_mp3_period(date_time)

        # assert
        period_exp = 'eYETS 2017/2018'
        self.assertEqual(period_exp, period_act)

    # Nominal voltage-dependent QH discharge reference criteria - 900 V
    def test_read_quench_heaters_reference_features_rb(self):
        # arrange
        circuit_type = 'RB'

        # act
        features_act = Reference.read_quench_heaters_reference_features(circuit_type)

        # assert
        features_ref = pd.DataFrame({'min': {'%CELL%:U_HDS_1:first': 780, '%CELL%:U_HDS_2:first': 780,
                                             '%CELL%:U_HDS_3:first': 780, '%CELL%:U_HDS_4:first': 780,
                                             '%CELL%:U_HDS_1:last20mean': 15, '%CELL%:U_HDS_2:last20mean': 15,
                                             '%CELL%:U_HDS_3:last20mean': 15, '%CELL%:U_HDS_4:last20mean': 15},
                                     'max': {'%CELL%:U_HDS_1:first': 980, '%CELL%:U_HDS_2:first': 980,
                                             '%CELL%:U_HDS_3:first': 980, '%CELL%:U_HDS_4:first': 980,
                                             '%CELL%:U_HDS_1:last20mean': 70, '%CELL%:U_HDS_2:last20mean': 70,
                                             '%CELL%:U_HDS_3:last20mean': 70, '%CELL%:U_HDS_4:last20mean': 70}})
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_features_rq(self):
        # arrange
        circuit_type = 'RQ'

        # act
        features_act = Reference.read_quench_heaters_reference_features(circuit_type)

        # assert
        features_ref = pd.DataFrame({'min': {'%CELL%:U_HDS_1:first': 810, '%CELL%:U_HDS_2:first': 810,
                                             '%CELL%:U_HDS_1:last20mean': 0, '%CELL%:U_HDS_2:last20mean': 0},
                                     'max': {'%CELL%:U_HDS_1:first': 1020, '%CELL%:U_HDS_2:first': 1020,
                                             '%CELL%:U_HDS_1:last20mean': 10, '%CELL%:U_HDS_2:last20mean': 10}})
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_features_it(self):
        # arrange
        circuit_type = 'IT'

        # act
        features_act = Reference.read_quench_heaters_reference_features(circuit_type)

        # assert
        features_ref = pd.DataFrame({'min': {'heat.%CIRCUIT%:U_HDS_1_Q1:first': 810,
                                             'heat.%CIRCUIT%:U_HDS_2_Q1:first': 810,
                                             'heat.%CIRCUIT%:U_HDS_1_Q2:first': 810,
                                             'heat.%CIRCUIT%:U_HDS_2_Q2:first': 810,
                                             'heat.%CIRCUIT%:U_HDS_3_Q2:first': 810,
                                             'heat.%CIRCUIT%:U_HDS_4_Q2:first': 810,
                                             'heat.%CIRCUIT%:U_HDS_1_Q3:first': 810,
                                             'heat.%CIRCUIT%:U_HDS_2_Q3:first': 810,
                                             'heat.%CIRCUIT%:U_HDS_1_Q1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_2_Q1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_1_Q2:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_2_Q2:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_3_Q2:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_4_Q2:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_1_Q3:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_2_Q3:last20mean': 0},
                                     'max': {'heat.%CIRCUIT%:U_HDS_1_Q1:first': 1000,
                                             'heat.%CIRCUIT%:U_HDS_2_Q1:first': 1000,
                                             'heat.%CIRCUIT%:U_HDS_1_Q2:first': 1000,
                                             'heat.%CIRCUIT%:U_HDS_2_Q2:first': 1000,
                                             'heat.%CIRCUIT%:U_HDS_3_Q2:first': 1000,
                                             'heat.%CIRCUIT%:U_HDS_4_Q2:first': 1000,
                                             'heat.%CIRCUIT%:U_HDS_1_Q3:first': 1000,
                                             'heat.%CIRCUIT%:U_HDS_2_Q3:first': 1000,
                                             'heat.%CIRCUIT%:U_HDS_1_Q1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_2_Q1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_1_Q2:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_2_Q2:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_3_Q2:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_4_Q2:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_1_Q3:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_2_Q3:last20mean': 10}})
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_features_ipd2(self):
        # arrange
        circuit_type = 'IPD2'

        # act
        features_act = Reference.read_quench_heaters_reference_features(circuit_type)

        # assert
        features_ref = pd.DataFrame({'min': {'heat.%CIRCUIT%:U_HDS_1:first': 810, 'heat.%CIRCUIT%:U_HDS_2:first': 810,
                                             'heat.%CIRCUIT%:U_HDS_1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_2:last20mean': 0},
                                     'max': {'heat.%CIRCUIT%:U_HDS_1:first': 1000, 'heat.%CIRCUIT%:U_HDS_2:first': 1000,
                                             'heat.%CIRCUIT%:U_HDS_1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_2:last20mean': 10}})
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_features_ipd2_b1b2(self):
        # arrange
        circuit_type = 'IPD2_B1B2'

        # act
        features_act = Reference.read_quench_heaters_reference_features(circuit_type)

        # assert
        features_ref = pd.DataFrame({'min': {'heat.%CIRCUIT%:U_HDS_1_B1:first': 810,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:first': 810,
                                             'heat.%CIRCUIT%:U_HDS_1_B1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:last20mean': 0},
                                     'max': {'heat.%CIRCUIT%:U_HDS_1_B1:first': 1000,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:first': 1000,
                                             'heat.%CIRCUIT%:U_HDS_1_B1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:last20mean': 10}})
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_features_ipq2(self):
        # arrange
        circuit_type = 'IPQ2'

        # act
        features_act = Reference.read_quench_heaters_reference_features(circuit_type)

        # assert
        features_ref = pd.DataFrame({'min': {'heat.%CIRCUIT%:U_HDS_1_B1:first': 810,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:first': 810,
                                             'heat.%CIRCUIT%:U_HDS_1_B1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:last20mean': 0},
                                     'max': {'heat.%CIRCUIT%:U_HDS_1_B1:first': 1000,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:first': 1000,
                                             'heat.%CIRCUIT%:U_HDS_1_B1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:last20mean': 10}})
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_features_ipq4(self):
        # arrange
        circuit_type = 'IPQ4'

        # act
        features_act = Reference.read_quench_heaters_reference_features(circuit_type)

        # assert
        features_ref = pd.DataFrame({'min': {'heat.%CIRCUIT%:U_HDS_1_B1:first': 810,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:first': 810,
                                             'heat.%CIRCUIT%:U_HDS_2_B1:first': 810,
                                             'heat.%CIRCUIT%:U_HDS_2_B2:first': 810,
                                             'heat.%CIRCUIT%:U_HDS_1_B1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_2_B1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_2_B2:last20mean': 0},
                                     'max': {'heat.%CIRCUIT%:U_HDS_1_B1:first': 1000,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:first': 1000,
                                             'heat.%CIRCUIT%:U_HDS_2_B1:first': 1000,
                                             'heat.%CIRCUIT%:U_HDS_2_B2:first': 1000,
                                             'heat.%CIRCUIT%:U_HDS_1_B1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_2_B1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_2_B2:last20mean': 10}})
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_features_ipq8(self):
        # arrange
        circuit_type = 'IPQ8'

        # act
        features_act = Reference.read_quench_heaters_reference_features(circuit_type)

        # assert
        features_ref = pd.DataFrame({'min': {'heat.%CIRCUIT%:U_HDS_1_B1:first': 810,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:first': 810,
                                             'heat.%CIRCUIT%:U_HDS_2_B1:first': 810,
                                             'heat.%CIRCUIT%:U_HDS_2_B2:first': 810,
                                             'heat.%CIRCUIT%:U_HDS_3_B1:first': 810,
                                             'heat.%CIRCUIT%:U_HDS_3_B2:first': 810,
                                             'heat.%CIRCUIT%:U_HDS_4_B1:first': 810,
                                             'heat.%CIRCUIT%:U_HDS_4_B2:first': 810,
                                             'heat.%CIRCUIT%:U_HDS_1_B1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_2_B1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_2_B2:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_3_B1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_3_B2:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_4_B1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_4_B2:last20mean': 0},
                                     'max': {'heat.%CIRCUIT%:U_HDS_1_B1:first': 1000,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:first': 1000,
                                             'heat.%CIRCUIT%:U_HDS_2_B1:first': 1000,
                                             'heat.%CIRCUIT%:U_HDS_2_B2:first': 1000,
                                             'heat.%CIRCUIT%:U_HDS_3_B1:first': 1000,
                                             'heat.%CIRCUIT%:U_HDS_3_B2:first': 1000,
                                             'heat.%CIRCUIT%:U_HDS_4_B1:first': 1000,
                                             'heat.%CIRCUIT%:U_HDS_4_B2:first': 1000,
                                             'heat.%CIRCUIT%:U_HDS_1_B1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_2_B1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_2_B2:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_3_B1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_3_B2:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_4_B1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_4_B2:last20mean': 10}})
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_differences_rb(self):
        # arrange
        circuit_type = 'RB'

        # act
        features_act = Reference.read_quench_heaters_reference_differences(circuit_type)

        # assert
        features_ref = pd.DataFrame({'diff': {'%CELL%:U_HDS_1:tau_charge': 0.003, '%CELL%:U_HDS_2:tau_charge': 0.003,
                                              '%CELL%:U_HDS_3:tau_charge': 0.003, '%CELL%:U_HDS_4:tau_charge': 0.003,
                                              '%CELL%:I_HDS_1:tau_charge': 0.003, '%CELL%:I_HDS_2:tau_charge': 0.003,
                                              '%CELL%:I_HDS_3:tau_charge': 0.003, '%CELL%:I_HDS_4:tau_charge': 0.003,
                                              '%CELL%:R_HDS_1:first20mean': 0.5, '%CELL%:R_HDS_2:first20mean': 0.5,
                                              '%CELL%:R_HDS_3:first20mean': 0.5, '%CELL%:R_HDS_4:first20mean': 0.5}})
        features_ref.sort_index(inplace=True)
        features_act.sort_index(inplace=True)
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_differences_rq(self):
        # arrange
        circuit_type = 'RQ'

        # act
        features_act = Reference.read_quench_heaters_reference_differences(circuit_type)

        # assert
        features_ref = pd.DataFrame({'diff': {'%CELL%:U_HDS_1:tau_charge': 0.003, '%CELL%:U_HDS_2:tau_charge': 0.003}})
        features_ref.sort_index(inplace=True)
        features_act.sort_index(inplace=True)
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_differences_it(self):
        # arrange
        circuit_type = 'IT'

        # act
        features_act = Reference.read_quench_heaters_reference_differences(circuit_type)

        # assert
        features_ref = pd.DataFrame({'diff': {'heat.%CIRCUIT%:U_HDS_1_Q1:tau_charge': 0.005,
                                              'heat.%CIRCUIT%:U_HDS_2_Q1:tau_charge': 0.005,
                                              'heat.%CIRCUIT%:U_HDS_1_Q2:tau_charge': 0.005,
                                              'heat.%CIRCUIT%:U_HDS_2_Q2:tau_charge': 0.005,
                                              'heat.%CIRCUIT%:U_HDS_3_Q2:tau_charge': 0.005,
                                              'heat.%CIRCUIT%:U_HDS_4_Q2:tau_charge': 0.005,
                                              'heat.%CIRCUIT%:U_HDS_1_Q3:tau_charge': 0.005,
                                              'heat.%CIRCUIT%:U_HDS_2_Q3:tau_charge': 0.005}})
        features_ref.sort_index(inplace=True)
        features_act.sort_index(inplace=True)
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_differences_ipd2(self):
        # arrange
        circuit_type = 'IPD2'

        # act
        features_act = Reference.read_quench_heaters_reference_differences(circuit_type)

        # assert
        features_ref = pd.DataFrame({'diff': {'heat.%CIRCUIT%:U_HDS_1:tau_charge': 0.005,
                                              'heat.%CIRCUIT%:U_HDS_2:tau_charge': 0.005}})
        features_ref.sort_index(inplace=True)
        features_act.sort_index(inplace=True)
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_differences_ipd2_b1b2(self):
        # arrange
        circuit_type = 'IPD2_B1B2'

        # act
        features_act = Reference.read_quench_heaters_reference_differences(circuit_type)

        # assert
        features_ref = pd.DataFrame({'diff': {'heat.%CIRCUIT%:U_HDS_1_B1:tau_charge': 0.005,
                                              'heat.%CIRCUIT%:U_HDS_1_B2:tau_charge': 0.005}})
        features_ref.sort_index(inplace=True)
        features_act.sort_index(inplace=True)
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_differences_ipq2(self):
        # arrange
        circuit_type = 'IPQ2'

        # act
        features_act = Reference.read_quench_heaters_reference_differences(circuit_type)

        # assert
        features_ref = pd.DataFrame({'diff': {'heat.%CIRCUIT%:U_HDS_1_B1:tau_charge': 0.005,
                                              'heat.%CIRCUIT%:U_HDS_1_B2:tau_charge': 0.005}})
        features_ref.sort_index(inplace=True)
        features_act.sort_index(inplace=True)
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_differences_ipq4(self):
        # arrange
        circuit_type = 'IPQ4'

        # act
        features_act = Reference.read_quench_heaters_reference_differences(circuit_type)

        # assert
        features_ref = pd.DataFrame({'diff': {'heat.%CIRCUIT%:U_HDS_1_B1:tau_charge': 0.005,
                                              'heat.%CIRCUIT%:U_HDS_1_B2:tau_charge': 0.005,
                                              'heat.%CIRCUIT%:U_HDS_2_B1:tau_charge': 0.005,
                                              'heat.%CIRCUIT%:U_HDS_2_B2:tau_charge': 0.005}})
        features_ref.sort_index(inplace=True)
        features_act.sort_index(inplace=True)
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_differences_ipq8(self):
        # arrange
        circuit_type = 'IPQ8'

        # act
        features_act = Reference.read_quench_heaters_reference_differences(circuit_type)

        # assert
        features_ref = pd.DataFrame({'diff': {'heat.%CIRCUIT%:U_HDS_1_B1:tau_charge': 0.005,
                                              'heat.%CIRCUIT%:U_HDS_1_B2:tau_charge': 0.005,
                                              'heat.%CIRCUIT%:U_HDS_2_B1:tau_charge': 0.005,
                                              'heat.%CIRCUIT%:U_HDS_2_B2:tau_charge': 0.005,
                                              'heat.%CIRCUIT%:U_HDS_3_B1:tau_charge': 0.005,
                                              'heat.%CIRCUIT%:U_HDS_3_B2:tau_charge': 0.005,
                                              'heat.%CIRCUIT%:U_HDS_4_B1:tau_charge': 0.005,
                                              'heat.%CIRCUIT%:U_HDS_4_B2:tau_charge': 0.005}})
        features_ref.sort_index(inplace=True)
        features_act.sort_index(inplace=True)
        pd.testing.assert_frame_equal(features_ref, features_act)

    # Nominal voltage-dependent QH discharge reference criteria - 300 V
    def test_read_quench_heaters_reference_features_rb_300v(self):
        # arrange
        circuit_type = 'RB'

        # act
        features_act = Reference.read_quench_heaters_reference_features(circuit_type, nominal_voltage=300)

        # assert
        features_ref = pd.DataFrame({'min': {'%CELL%:U_HDS_1:first': 250, '%CELL%:U_HDS_2:first': 250,
                                             '%CELL%:U_HDS_3:first': 250, '%CELL%:U_HDS_4:first': 250,
                                             '%CELL%:U_HDS_1:last20mean': 0, '%CELL%:U_HDS_2:last20mean': 0,
                                             '%CELL%:U_HDS_3:last20mean': 0, '%CELL%:U_HDS_4:last20mean': 0},
                                     'max': {'%CELL%:U_HDS_1:first': 350, '%CELL%:U_HDS_2:first': 350,
                                             '%CELL%:U_HDS_3:first': 350, '%CELL%:U_HDS_4:first': 350,
                                             '%CELL%:U_HDS_1:last20mean': 10, '%CELL%:U_HDS_2:last20mean': 10,
                                             '%CELL%:U_HDS_3:last20mean': 10, '%CELL%:U_HDS_4:last20mean': 10}})
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_features_rq_300v(self):
        # arrange
        circuit_type = 'RQ'

        # act
        features_act = Reference.read_quench_heaters_reference_features(circuit_type, nominal_voltage=300)

        # assert
        features_ref = pd.DataFrame({'min': {'%CELL%:U_HDS_1:first': 250, '%CELL%:U_HDS_2:first': 250,
                                             '%CELL%:U_HDS_1:last20mean': 0, '%CELL%:U_HDS_2:last20mean': 0},
                                     'max': {'%CELL%:U_HDS_1:first': 350, '%CELL%:U_HDS_2:first': 350,
                                             '%CELL%:U_HDS_1:last20mean': 10, '%CELL%:U_HDS_2:last20mean': 10}})
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_features_it_300v(self):
        # arrange
        circuit_type = 'IT'

        # act
        features_act = Reference.read_quench_heaters_reference_features(circuit_type, nominal_voltage=300)

        # assert
        features_ref = pd.DataFrame({'min': {'heat.%CIRCUIT%:U_HDS_1_Q1:first': 250,
                                             'heat.%CIRCUIT%:U_HDS_2_Q1:first': 250,
                                             'heat.%CIRCUIT%:U_HDS_1_Q2:first': 250,
                                             'heat.%CIRCUIT%:U_HDS_2_Q2:first': 250,
                                             'heat.%CIRCUIT%:U_HDS_3_Q2:first': 250,
                                             'heat.%CIRCUIT%:U_HDS_4_Q2:first': 250,
                                             'heat.%CIRCUIT%:U_HDS_1_Q3:first': 250,
                                             'heat.%CIRCUIT%:U_HDS_2_Q3:first': 250,
                                             'heat.%CIRCUIT%:U_HDS_1_Q1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_2_Q1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_1_Q2:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_2_Q2:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_3_Q2:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_4_Q2:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_1_Q3:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_2_Q3:last20mean': 0},
                                     'max': {'heat.%CIRCUIT%:U_HDS_1_Q1:first': 350,
                                             'heat.%CIRCUIT%:U_HDS_2_Q1:first': 350,
                                             'heat.%CIRCUIT%:U_HDS_1_Q2:first': 350,
                                             'heat.%CIRCUIT%:U_HDS_2_Q2:first': 350,
                                             'heat.%CIRCUIT%:U_HDS_3_Q2:first': 350,
                                             'heat.%CIRCUIT%:U_HDS_4_Q2:first': 350,
                                             'heat.%CIRCUIT%:U_HDS_1_Q3:first': 350,
                                             'heat.%CIRCUIT%:U_HDS_2_Q3:first': 350,
                                             'heat.%CIRCUIT%:U_HDS_1_Q1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_2_Q1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_1_Q2:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_2_Q2:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_3_Q2:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_4_Q2:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_1_Q3:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_2_Q3:last20mean': 10}})
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_features_ipd2_300v(self):
        # arrange
        circuit_type = 'IPD2'

        # act
        features_act = Reference.read_quench_heaters_reference_features(circuit_type, nominal_voltage=300)

        # assert
        features_ref = pd.DataFrame({'min': {'heat.%CIRCUIT%:U_HDS_1:first': 250, 'heat.%CIRCUIT%:U_HDS_2:first': 250,
                                             'heat.%CIRCUIT%:U_HDS_1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_2:last20mean': 0},
                                     'max': {'heat.%CIRCUIT%:U_HDS_1:first': 350, 'heat.%CIRCUIT%:U_HDS_2:first': 350,
                                             'heat.%CIRCUIT%:U_HDS_1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_2:last20mean': 10}})
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_features_ipd2_b1b2_300v(self):
        # arrange
        circuit_type = 'IPD2_B1B2'

        # act
        features_act = Reference.read_quench_heaters_reference_features(circuit_type, nominal_voltage=300)

        # assert
        features_ref = pd.DataFrame({'min': {'heat.%CIRCUIT%:U_HDS_1_B1:first': 250,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:first': 250,
                                             'heat.%CIRCUIT%:U_HDS_1_B1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:last20mean': 0},
                                     'max': {'heat.%CIRCUIT%:U_HDS_1_B1:first': 350,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:first': 350,
                                             'heat.%CIRCUIT%:U_HDS_1_B1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:last20mean': 10}})
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_features_ipq2_300v(self):
        # arrange
        circuit_type = 'IPQ2'

        # act
        features_act = Reference.read_quench_heaters_reference_features(circuit_type, nominal_voltage=300)

        # assert
        features_ref = pd.DataFrame({'min': {'heat.%CIRCUIT%:U_HDS_1_B1:first': 250,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:first': 250,
                                             'heat.%CIRCUIT%:U_HDS_1_B1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:last20mean': 0},
                                     'max': {'heat.%CIRCUIT%:U_HDS_1_B1:first': 350,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:first': 350,
                                             'heat.%CIRCUIT%:U_HDS_1_B1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:last20mean': 10}})
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_features_ipq4_300v(self):
        # arrange
        circuit_type = 'IPQ4'

        # act
        features_act = Reference.read_quench_heaters_reference_features(circuit_type, nominal_voltage=300)

        # assert
        features_ref = pd.DataFrame({'min': {'heat.%CIRCUIT%:U_HDS_1_B1:first': 250,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:first': 250,
                                             'heat.%CIRCUIT%:U_HDS_2_B1:first': 250,
                                             'heat.%CIRCUIT%:U_HDS_2_B2:first': 250,
                                             'heat.%CIRCUIT%:U_HDS_1_B1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_2_B1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_2_B2:last20mean': 0},
                                     'max': {'heat.%CIRCUIT%:U_HDS_1_B1:first': 350,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:first': 350,
                                             'heat.%CIRCUIT%:U_HDS_2_B1:first': 350,
                                             'heat.%CIRCUIT%:U_HDS_2_B2:first': 350,
                                             'heat.%CIRCUIT%:U_HDS_1_B1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_2_B1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_2_B2:last20mean': 10}})
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_features_ipq8_300v(self):
        # arrange
        circuit_type = 'IPQ8'

        # act
        features_act = Reference.read_quench_heaters_reference_features(circuit_type, nominal_voltage=300)

        # assert
        features_ref = pd.DataFrame({'min': {'heat.%CIRCUIT%:U_HDS_1_B1:first': 250,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:first': 250,
                                             'heat.%CIRCUIT%:U_HDS_2_B1:first': 250,
                                             'heat.%CIRCUIT%:U_HDS_2_B2:first': 250,
                                             'heat.%CIRCUIT%:U_HDS_3_B1:first': 250,
                                             'heat.%CIRCUIT%:U_HDS_3_B2:first': 250,
                                             'heat.%CIRCUIT%:U_HDS_4_B1:first': 250,
                                             'heat.%CIRCUIT%:U_HDS_4_B2:first': 250,
                                             'heat.%CIRCUIT%:U_HDS_1_B1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_2_B1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_2_B2:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_3_B1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_3_B2:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_4_B1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_4_B2:last20mean': 0},
                                     'max': {'heat.%CIRCUIT%:U_HDS_1_B1:first': 350,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:first': 350,
                                             'heat.%CIRCUIT%:U_HDS_2_B1:first': 350,
                                             'heat.%CIRCUIT%:U_HDS_2_B2:first': 350,
                                             'heat.%CIRCUIT%:U_HDS_3_B1:first': 350,
                                             'heat.%CIRCUIT%:U_HDS_3_B2:first': 350,
                                             'heat.%CIRCUIT%:U_HDS_4_B1:first': 350,
                                             'heat.%CIRCUIT%:U_HDS_4_B2:first': 350,
                                             'heat.%CIRCUIT%:U_HDS_1_B1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_2_B1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_2_B2:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_3_B1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_3_B2:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_4_B1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_4_B2:last20mean': 10}})
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_differences_rb_300v(self):
        # arrange
        circuit_type = 'RB'

        # act
        features_act = Reference.read_quench_heaters_reference_differences(circuit_type, nominal_voltage=300)

        # assert
        features_ref = pd.DataFrame({'diff': {'%CELL%:U_HDS_1:tau_charge': 0.01, '%CELL%:U_HDS_2:tau_charge': 0.01,
                                              '%CELL%:U_HDS_3:tau_charge': 0.01, '%CELL%:U_HDS_4:tau_charge': 0.01,
                                              '%CELL%:I_HDS_1:tau_charge': 0.01, '%CELL%:I_HDS_2:tau_charge': 0.01,
                                              '%CELL%:I_HDS_3:tau_charge': 0.01, '%CELL%:I_HDS_4:tau_charge': 0.01,
                                              '%CELL%:R_HDS_1:first20mean': 0.5, '%CELL%:R_HDS_2:first20mean': 0.5,
                                              '%CELL%:R_HDS_3:first20mean': 0.5, '%CELL%:R_HDS_4:first20mean': 0.5}})
        features_ref.sort_index(inplace=True)
        features_act.sort_index(inplace=True)
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_differences_rq_300v(self):
        # arrange
        circuit_type = 'RQ'

        # act
        features_act = Reference.read_quench_heaters_reference_differences(circuit_type, nominal_voltage=300)

        # assert
        features_ref = pd.DataFrame({'diff': {'%CELL%:U_HDS_1:tau_charge': 0.01, '%CELL%:U_HDS_2:tau_charge': 0.01}})
        features_ref.sort_index(inplace=True)
        features_act.sort_index(inplace=True)
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_differences_it_300v(self):
        # arrange
        circuit_type = 'IT'

        # act
        features_act = Reference.read_quench_heaters_reference_differences(circuit_type, nominal_voltage=300)

        # assert
        features_ref = pd.DataFrame({'diff': {'heat.%CIRCUIT%:U_HDS_1_Q1:tau_charge': 0.01,
                                              'heat.%CIRCUIT%:U_HDS_2_Q1:tau_charge': 0.01,
                                              'heat.%CIRCUIT%:U_HDS_1_Q2:tau_charge': 0.01,
                                              'heat.%CIRCUIT%:U_HDS_2_Q2:tau_charge': 0.01,
                                              'heat.%CIRCUIT%:U_HDS_3_Q2:tau_charge': 0.01,
                                              'heat.%CIRCUIT%:U_HDS_4_Q2:tau_charge': 0.01,
                                              'heat.%CIRCUIT%:U_HDS_1_Q3:tau_charge': 0.01,
                                              'heat.%CIRCUIT%:U_HDS_2_Q3:tau_charge': 0.01}})
        features_ref.sort_index(inplace=True)
        features_act.sort_index(inplace=True)
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_differences_ipd2_300v(self):
        # arrange
        circuit_type = 'IPD2'

        # act
        features_act = Reference.read_quench_heaters_reference_differences(circuit_type, nominal_voltage=300)

        # assert
        features_ref = pd.DataFrame({'diff': {'heat.%CIRCUIT%:U_HDS_1:tau_charge': 0.01,
                                              'heat.%CIRCUIT%:U_HDS_2:tau_charge': 0.01}})
        features_ref.sort_index(inplace=True)
        features_act.sort_index(inplace=True)
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_differences_ipd2_b1b2_300v(self):
        # arrange
        circuit_type = 'IPD2_B1B2'

        # act
        features_act = Reference.read_quench_heaters_reference_differences(circuit_type, nominal_voltage=300)

        # assert
        features_ref = pd.DataFrame({'diff': {'heat.%CIRCUIT%:U_HDS_1_B1:tau_charge': 0.01,
                                              'heat.%CIRCUIT%:U_HDS_1_B2:tau_charge': 0.01}})
        features_ref.sort_index(inplace=True)
        features_act.sort_index(inplace=True)
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_differences_ipq2_300v(self):
        # arrange
        circuit_type = 'IPQ2'

        # act
        features_act = Reference.read_quench_heaters_reference_differences(circuit_type, nominal_voltage=300)

        # assert
        features_ref = pd.DataFrame({'diff': {'heat.%CIRCUIT%:U_HDS_1_B1:tau_charge': 0.01,
                                              'heat.%CIRCUIT%:U_HDS_1_B2:tau_charge': 0.01}})
        features_ref.sort_index(inplace=True)
        features_act.sort_index(inplace=True)
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_differences_ipq4_300v(self):
        # arrange
        circuit_type = 'IPQ4'

        # act
        features_act = Reference.read_quench_heaters_reference_differences(circuit_type, nominal_voltage=300)

        # assert
        features_ref = pd.DataFrame({'diff': {'heat.%CIRCUIT%:U_HDS_1_B1:tau_charge': 0.01,
                                              'heat.%CIRCUIT%:U_HDS_1_B2:tau_charge': 0.01,
                                              'heat.%CIRCUIT%:U_HDS_2_B1:tau_charge': 0.01,
                                              'heat.%CIRCUIT%:U_HDS_2_B2:tau_charge': 0.01}})
        features_ref.sort_index(inplace=True)
        features_act.sort_index(inplace=True)
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_differences_ipq8_300v(self):
        # arrange
        circuit_type = 'IPQ8'

        # act
        features_act = Reference.read_quench_heaters_reference_differences(circuit_type, nominal_voltage=300)

        # assert
        features_ref = pd.DataFrame({'diff': {'heat.%CIRCUIT%:U_HDS_1_B1:tau_charge': 0.01,
                                              'heat.%CIRCUIT%:U_HDS_1_B2:tau_charge': 0.01,
                                              'heat.%CIRCUIT%:U_HDS_2_B1:tau_charge': 0.01,
                                              'heat.%CIRCUIT%:U_HDS_2_B2:tau_charge': 0.01,
                                              'heat.%CIRCUIT%:U_HDS_3_B1:tau_charge': 0.01,
                                              'heat.%CIRCUIT%:U_HDS_3_B2:tau_charge': 0.01,
                                              'heat.%CIRCUIT%:U_HDS_4_B1:tau_charge': 0.01,
                                              'heat.%CIRCUIT%:U_HDS_4_B2:tau_charge': 0.01}})
        features_ref.sort_index(inplace=True)
        features_act.sort_index(inplace=True)
        pd.testing.assert_frame_equal(features_ref, features_act)

    # Nominal voltage-dependent QH discharge reference criteria - 10 V
    def test_read_quench_heaters_reference_features_rb_10v(self):
        # arrange
        circuit_type = 'RB'

        # act
        features_act = Reference.read_quench_heaters_reference_features(circuit_type, nominal_voltage=10)

        # assert
        features_ref = pd.DataFrame({'min': {'%CELL%:U_HDS_1:first': 5, '%CELL%:U_HDS_2:first': 5,
                                             '%CELL%:U_HDS_3:first': 5, '%CELL%:U_HDS_4:first': 5,
                                             '%CELL%:U_HDS_1:last20mean': 0, '%CELL%:U_HDS_2:last20mean': 0,
                                             '%CELL%:U_HDS_3:last20mean': 0, '%CELL%:U_HDS_4:last20mean': 0},
                                     'max': {'%CELL%:U_HDS_1:first': 15, '%CELL%:U_HDS_2:first': 15,
                                             '%CELL%:U_HDS_3:first': 15, '%CELL%:U_HDS_4:first': 15,
                                             '%CELL%:U_HDS_1:last20mean': 10, '%CELL%:U_HDS_2:last20mean': 10,
                                             '%CELL%:U_HDS_3:last20mean': 10, '%CELL%:U_HDS_4:last20mean': 10}})
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_features_rq_10v(self):
        # arrange
        circuit_type = 'RQ'

        # act
        features_act = Reference.read_quench_heaters_reference_features(circuit_type, nominal_voltage=10)

        # assert
        features_ref = pd.DataFrame({'min': {'%CELL%:U_HDS_1:first': 5, '%CELL%:U_HDS_2:first': 5,
                                             '%CELL%:U_HDS_1:last20mean': 0, '%CELL%:U_HDS_2:last20mean': 0},
                                     'max': {'%CELL%:U_HDS_1:first': 15, '%CELL%:U_HDS_2:first': 15,
                                             '%CELL%:U_HDS_1:last20mean': 10, '%CELL%:U_HDS_2:last20mean': 10}})
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_features_it_10v(self):
        # arrange
        circuit_type = 'IT'

        # act
        features_act = Reference.read_quench_heaters_reference_features(circuit_type, nominal_voltage=10)

        # assert
        features_ref = pd.DataFrame({'min': {'heat.%CIRCUIT%:U_HDS_1_Q1:first': 5, 'heat.%CIRCUIT%:U_HDS_2_Q1:first': 5,
                                             'heat.%CIRCUIT%:U_HDS_1_Q2:first': 5, 'heat.%CIRCUIT%:U_HDS_2_Q2:first': 5,
                                             'heat.%CIRCUIT%:U_HDS_3_Q2:first': 5, 'heat.%CIRCUIT%:U_HDS_4_Q2:first': 5,
                                             'heat.%CIRCUIT%:U_HDS_1_Q3:first': 5, 'heat.%CIRCUIT%:U_HDS_2_Q3:first': 5,
                                             'heat.%CIRCUIT%:U_HDS_1_Q1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_2_Q1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_1_Q2:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_2_Q2:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_3_Q2:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_4_Q2:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_1_Q3:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_2_Q3:last20mean': 0},
                                     'max': {'heat.%CIRCUIT%:U_HDS_1_Q1:first': 15,
                                             'heat.%CIRCUIT%:U_HDS_2_Q1:first': 15,
                                             'heat.%CIRCUIT%:U_HDS_1_Q2:first': 15,
                                             'heat.%CIRCUIT%:U_HDS_2_Q2:first': 15,
                                             'heat.%CIRCUIT%:U_HDS_3_Q2:first': 15,
                                             'heat.%CIRCUIT%:U_HDS_4_Q2:first': 15,
                                             'heat.%CIRCUIT%:U_HDS_1_Q3:first': 15,
                                             'heat.%CIRCUIT%:U_HDS_2_Q3:first': 15,
                                             'heat.%CIRCUIT%:U_HDS_1_Q1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_2_Q1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_1_Q2:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_2_Q2:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_3_Q2:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_4_Q2:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_1_Q3:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_2_Q3:last20mean': 10}})
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_features_ipd2_10v(self):
        # arrange
        circuit_type = 'IPD2'

        # act
        features_act = Reference.read_quench_heaters_reference_features(circuit_type, nominal_voltage=10)

        # assert
        features_ref = pd.DataFrame({'min': {'heat.%CIRCUIT%:U_HDS_1:first': 5, 'heat.%CIRCUIT%:U_HDS_2:first': 5,
                                             'heat.%CIRCUIT%:U_HDS_1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_2:last20mean': 0},
                                     'max': {'heat.%CIRCUIT%:U_HDS_1:first': 15, 'heat.%CIRCUIT%:U_HDS_2:first': 15,
                                             'heat.%CIRCUIT%:U_HDS_1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_2:last20mean': 10}})
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_features_ipd2_b1b2_10v(self):
        # arrange
        circuit_type = 'IPD2_B1B2'

        # act
        features_act = Reference.read_quench_heaters_reference_features(circuit_type, nominal_voltage=10)

        # assert
        features_ref = pd.DataFrame({'min': {'heat.%CIRCUIT%:U_HDS_1_B1:first': 5, 'heat.%CIRCUIT%:U_HDS_1_B2:first': 5,
                                             'heat.%CIRCUIT%:U_HDS_1_B1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:last20mean': 0},
                                     'max': {'heat.%CIRCUIT%:U_HDS_1_B1:first': 15,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:first': 15,
                                             'heat.%CIRCUIT%:U_HDS_1_B1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:last20mean': 10}})
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_features_ipq2_10v(self):
        # arrange
        circuit_type = 'IPQ2'

        # act
        features_act = Reference.read_quench_heaters_reference_features(circuit_type, nominal_voltage=10)

        # assert
        features_ref = pd.DataFrame({'min': {'heat.%CIRCUIT%:U_HDS_1_B1:first': 5, 'heat.%CIRCUIT%:U_HDS_1_B2:first': 5,
                                             'heat.%CIRCUIT%:U_HDS_1_B1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:last20mean': 0},
                                     'max': {'heat.%CIRCUIT%:U_HDS_1_B1:first': 15,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:first': 15,
                                             'heat.%CIRCUIT%:U_HDS_1_B1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:last20mean': 10}})
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_features_ipq4_10v(self):
        # arrange
        circuit_type = 'IPQ4'

        # act
        features_act = Reference.read_quench_heaters_reference_features(circuit_type, nominal_voltage=10)

        # assert
        features_ref = pd.DataFrame({'min': {'heat.%CIRCUIT%:U_HDS_1_B1:first': 5, 'heat.%CIRCUIT%:U_HDS_1_B2:first': 5,
                                             'heat.%CIRCUIT%:U_HDS_2_B1:first': 5, 'heat.%CIRCUIT%:U_HDS_2_B2:first': 5,
                                             'heat.%CIRCUIT%:U_HDS_1_B1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_2_B1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_2_B2:last20mean': 0},
                                     'max': {'heat.%CIRCUIT%:U_HDS_1_B1:first': 15,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:first': 15,
                                             'heat.%CIRCUIT%:U_HDS_2_B1:first': 15,
                                             'heat.%CIRCUIT%:U_HDS_2_B2:first': 15,
                                             'heat.%CIRCUIT%:U_HDS_1_B1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_2_B1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_2_B2:last20mean': 10}})
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_features_ipq8_10v(self):
        # arrange
        circuit_type = 'IPQ8'

        # act
        features_act = Reference.read_quench_heaters_reference_features(circuit_type, nominal_voltage=10)

        # assert
        features_ref = pd.DataFrame({'min': {'heat.%CIRCUIT%:U_HDS_1_B1:first': 5, 'heat.%CIRCUIT%:U_HDS_1_B2:first': 5,
                                             'heat.%CIRCUIT%:U_HDS_2_B1:first': 5, 'heat.%CIRCUIT%:U_HDS_2_B2:first': 5,
                                             'heat.%CIRCUIT%:U_HDS_3_B1:first': 5, 'heat.%CIRCUIT%:U_HDS_3_B2:first': 5,
                                             'heat.%CIRCUIT%:U_HDS_4_B1:first': 5, 'heat.%CIRCUIT%:U_HDS_4_B2:first': 5,
                                             'heat.%CIRCUIT%:U_HDS_1_B1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_2_B1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_2_B2:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_3_B1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_3_B2:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_4_B1:last20mean': 0,
                                             'heat.%CIRCUIT%:U_HDS_4_B2:last20mean': 0},
                                     'max': {'heat.%CIRCUIT%:U_HDS_1_B1:first': 15,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:first': 15,
                                             'heat.%CIRCUIT%:U_HDS_2_B1:first': 15,
                                             'heat.%CIRCUIT%:U_HDS_2_B2:first': 15,
                                             'heat.%CIRCUIT%:U_HDS_3_B1:first': 15,
                                             'heat.%CIRCUIT%:U_HDS_3_B2:first': 15,
                                             'heat.%CIRCUIT%:U_HDS_4_B1:first': 15,
                                             'heat.%CIRCUIT%:U_HDS_4_B2:first': 15,
                                             'heat.%CIRCUIT%:U_HDS_1_B1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_1_B2:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_2_B1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_2_B2:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_3_B1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_3_B2:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_4_B1:last20mean': 10,
                                             'heat.%CIRCUIT%:U_HDS_4_B2:last20mean': 10}})
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_differences_rb_10v(self):
        # arrange
        circuit_type = 'RB'

        # act
        features_act = Reference.read_quench_heaters_reference_differences(circuit_type, nominal_voltage=10)

        # assert
        features_ref = pd.DataFrame({'diff': {'%CELL%:U_HDS_1:tau_charge': 0.02, '%CELL%:U_HDS_2:tau_charge': 0.02,
                                              '%CELL%:U_HDS_3:tau_charge': 0.02, '%CELL%:U_HDS_4:tau_charge': 0.02,
                                              '%CELL%:I_HDS_1:tau_charge': 0.02, '%CELL%:I_HDS_2:tau_charge': 0.02,
                                              '%CELL%:I_HDS_3:tau_charge': 0.02, '%CELL%:I_HDS_4:tau_charge': 0.02,
                                              '%CELL%:R_HDS_1:first20mean': 0.5, '%CELL%:R_HDS_2:first20mean': 0.5,
                                              '%CELL%:R_HDS_3:first20mean': 0.5, '%CELL%:R_HDS_4:first20mean': 0.5}})
        features_ref.sort_index(inplace=True)
        features_act.sort_index(inplace=True)
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_differences_rq_10v(self):
        # arrange
        circuit_type = 'RQ'

        # act
        features_act = Reference.read_quench_heaters_reference_differences(circuit_type, nominal_voltage=10)

        # assert
        features_ref = pd.DataFrame({'diff': {'%CELL%:U_HDS_1:tau_charge': 0.02, '%CELL%:U_HDS_2:tau_charge': 0.02}})
        features_ref.sort_index(inplace=True)
        features_act.sort_index(inplace=True)
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_differences_it_10v(self):
        # arrange
        circuit_type = 'IT'

        # act
        features_act = Reference.read_quench_heaters_reference_differences(circuit_type, nominal_voltage=10)

        # assert
        features_ref = pd.DataFrame({'diff': {'heat.%CIRCUIT%:U_HDS_1_Q1:tau_charge': 0.02,
                                              'heat.%CIRCUIT%:U_HDS_2_Q1:tau_charge': 0.02,
                                              'heat.%CIRCUIT%:U_HDS_1_Q2:tau_charge': 0.02,
                                              'heat.%CIRCUIT%:U_HDS_2_Q2:tau_charge': 0.02,
                                              'heat.%CIRCUIT%:U_HDS_3_Q2:tau_charge': 0.02,
                                              'heat.%CIRCUIT%:U_HDS_4_Q2:tau_charge': 0.02,
                                              'heat.%CIRCUIT%:U_HDS_1_Q3:tau_charge': 0.02,
                                              'heat.%CIRCUIT%:U_HDS_2_Q3:tau_charge': 0.02}})
        features_ref.sort_index(inplace=True)
        features_act.sort_index(inplace=True)
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_differences_ipd2_10v(self):
        # arrange
        circuit_type = 'IPD2'

        # act
        features_act = Reference.read_quench_heaters_reference_differences(circuit_type, nominal_voltage=10)

        # assert
        features_ref = pd.DataFrame({'diff': {'heat.%CIRCUIT%:U_HDS_1:tau_charge': 0.02,
                                              'heat.%CIRCUIT%:U_HDS_2:tau_charge': 0.02}})
        features_ref.sort_index(inplace=True)
        features_act.sort_index(inplace=True)
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_differences_ipd2_b1b2_10v(self):
        # arrange
        circuit_type = 'IPD2_B1B2'

        # act
        features_act = Reference.read_quench_heaters_reference_differences(circuit_type, nominal_voltage=10)

        # assert
        features_ref = pd.DataFrame({'diff': {'heat.%CIRCUIT%:U_HDS_1_B1:tau_charge': 0.02,
                                              'heat.%CIRCUIT%:U_HDS_1_B2:tau_charge': 0.02}})
        features_ref.sort_index(inplace=True)
        features_act.sort_index(inplace=True)
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_differences_ipq2_10v(self):
        # arrange
        circuit_type = 'IPQ2'

        # act
        features_act = Reference.read_quench_heaters_reference_differences(circuit_type, nominal_voltage=10)

        # assert
        features_ref = pd.DataFrame({'diff': {'heat.%CIRCUIT%:U_HDS_1_B1:tau_charge': 0.02,
                                              'heat.%CIRCUIT%:U_HDS_1_B2:tau_charge': 0.02}})
        features_ref.sort_index(inplace=True)
        features_act.sort_index(inplace=True)
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_differences_ipq4_10v(self):
        # arrange
        circuit_type = 'IPQ4'

        # act
        features_act = Reference.read_quench_heaters_reference_differences(circuit_type, nominal_voltage=10)

        # assert
        features_ref = pd.DataFrame({'diff': {'heat.%CIRCUIT%:U_HDS_1_B1:tau_charge': 0.02,
                                              'heat.%CIRCUIT%:U_HDS_1_B2:tau_charge': 0.02,
                                              'heat.%CIRCUIT%:U_HDS_2_B1:tau_charge': 0.02,
                                              'heat.%CIRCUIT%:U_HDS_2_B2:tau_charge': 0.02}})
        features_ref.sort_index(inplace=True)
        features_act.sort_index(inplace=True)
        pd.testing.assert_frame_equal(features_ref, features_act)

    def test_read_quench_heaters_reference_differences_ipq8_10v(self):
        # arrange
        circuit_type = 'IPQ8'

        # act
        features_act = Reference.read_quench_heaters_reference_differences(circuit_type, nominal_voltage=10)

        # assert
        features_ref = pd.DataFrame({'diff': {'heat.%CIRCUIT%:U_HDS_1_B1:tau_charge': 0.02,
                                              'heat.%CIRCUIT%:U_HDS_1_B2:tau_charge': 0.02,
                                              'heat.%CIRCUIT%:U_HDS_2_B1:tau_charge': 0.02,
                                              'heat.%CIRCUIT%:U_HDS_2_B2:tau_charge': 0.02,
                                              'heat.%CIRCUIT%:U_HDS_3_B1:tau_charge': 0.02,
                                              'heat.%CIRCUIT%:U_HDS_3_B2:tau_charge': 0.02,
                                              'heat.%CIRCUIT%:U_HDS_4_B1:tau_charge': 0.02,
                                              'heat.%CIRCUIT%:U_HDS_4_B2:tau_charge': 0.02}})
        features_ref.sort_index(inplace=True)
        features_act.sort_index(inplace=True)
        pd.testing.assert_frame_equal(features_ref, features_act)

    # Support to several reference QH discharges to account for magnet replacement over time
    def test_get_quench_heater_reference_discharge_b24l5(self):
        """This test checks the retrieval of a QH reference for no input timestamp"""
        # arrange
        circuit_type = 'RB'
        magnet_name = 'B24L5'

        # act
        timestamp_act = Reference.get_quench_heater_reference_discharge(circuit_type, magnet_name)

        # assert
        timestamp_ref = 1424018504014000000
        self.assertEqual(timestamp_ref, timestamp_act)

    def test_get_quench_heater_reference_discharge_b24l5_equal_to_first_reference(self):
        """This test checks the retrieval of a QH reference for no input timestamp"""
        # arrange
        circuit_type = 'RB'
        magnet_name = 'B24L5'
        timestamp = 1424018504014000000

        # act
        timestamp_act = Reference.get_quench_heater_reference_discharge(circuit_type, magnet_name, timestamp)

        # assert
        timestamp_ref = 1424018504014000000
        self.assertEqual(timestamp_ref, timestamp_act)

    def test_get_quench_heater_reference_discharge_b24l5_before_reference(self):
        """This test checks the retrieval of a QH reference for input timestamp before the first reference.
        A warning should be displayed and the first timestamp returned"""
        # arrange
        circuit_type = 'RB'
        magnet_name = 'B24L5'
        timestamp = 1424018504014000000 - 10

        # act
        # assert
        timestamp_ref = 1424018504014000000
        with warnings.catch_warnings(record=True) as w:
            timestamp_act = Reference.get_quench_heater_reference_discharge(circuit_type, magnet_name, timestamp)
            self.assertEqual('The reference QH discharge is older than the given timestamp!', str(w[0].message))
            self.assertEqual(timestamp_ref, timestamp_act)

    def test_get_quench_heater_reference_discharge_b24l5_after_first_reference(self):
        """This test checks the retrieval of a QH reference for input timestamp after the first reference.
        The first timestamp returned"""
        # arrange
        circuit_type = 'RB'
        magnet_name = 'B24L5'
        timestamp = 1424018504014000000 + 10

        # act
        timestamp_act = Reference.get_quench_heater_reference_discharge(circuit_type, magnet_name, timestamp)

        # assert
        timestamp_ref = 1424018504014000000
        self.assertEqual(timestamp_ref, timestamp_act)

    def test_get_quench_heater_reference_discharge_b24l5_equal_to_last_reference(self):
        """This test checks the retrieval of a QH reference for input timestamp after the first reference.
        The first timestamp returned"""
        # arrange
        circuit_type = 'RB'
        magnet_name = 'B24L5'
        timestamp = 1611330989888000000

        # act
        timestamp_act = Reference.get_quench_heater_reference_discharge(circuit_type, magnet_name, timestamp)

        # assert
        timestamp_ref = 1611330989888000000
        self.assertEqual(timestamp_ref, timestamp_act)

    def test_get_quench_heater_reference_discharge_b24l5_after_last_reference(self):
        """This test checks the retrieval of a QH reference for input timestamp after the first reference.
        The first timestamp returned"""
        # arrange
        circuit_type = 'RB'
        magnet_name = 'B24L5'
        timestamp = 1611330989888000000 + 10

        # act
        timestamp_act = Reference.get_quench_heater_reference_discharge(circuit_type, magnet_name, timestamp)

        # assert
        timestamp_ref = 1611330989888000000
        self.assertEqual(timestamp_ref, timestamp_act)
